@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2011-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2011-2015 -- leonerd@leonerd.org.uk
 
 package Future;
 
@@ -9,7 +9,7 @@ use strict;
 use warnings;
 no warnings 'recursion'; # Disable the "deep recursion" warning
 
-our $VERSION = '0.30';
+our $VERSION = '0.31';
 
 use Carp qw(); # don't import croak
 use Scalar::Util qw( weaken blessed reftype );
@@ -138,8 +138,6 @@ or cancelled. By enabling debug tracing of objects, this fact can be checked.
 If a future object is destroyed without having been completed or cancelled, a
 warning message is printed.
 
-This feature is enabled by setting an environment variable called
-C<PERL_FUTURE_DEBUG> to some true value.
 
  $ PERL_FUTURE_DEBUG=1 perl -MFuture -E 'my $f = Future->new'
  Future=HASH(0xaa61f8) was constructed at -e line 1 and was lost near -e line 0 before it was ready.
@@ -161,6 +159,18 @@ line following the line calling that function.
  Future=HASH(0x14a2220) was constructed at - line 2 and was lost near - line 6 before it was ready.
  Finished
 
+A warning is also printed in debug mode if a C<Future> object is destroyed
+that completed with a failure, but the object believes that failure has not
+been reported anywhere.
+
+ $ PERL_FUTURE_DEBUG=1 perl -Mblib -MFuture -E 'my $f = Future->fail("Oops")'
+ Future=HASH(0xac98f8) was constructed at -e line 1 and was lost near -e line 0 with an unreported failure of: Oops
+
+Such a failure is considered reported if the C<get> or C<failure> methods are
+called on it, or it had at least one C<on_ready> or C<on_fail> callback, or
+its failure is propagated to another C<Future> instance (by a sequencing or
+converging method).
+
 =cut
 
 =head1 CONSTRUCTORS
@@ -244,19 +254,26 @@ sub new
 my $GLOBAL_END;
 END { $GLOBAL_END = 1; }
 
-*DESTROY = sub {
+sub DESTROY_debug {
    my $self = shift;
    return if $GLOBAL_END;
-
-   return if $self->{ready};
+   return if $self->{ready} and ( $self->{reported} or !$self->{failure} );
 
    my $lost_at = join " line ", (caller)[1,2];
    # We can't actually know the real line where the last reference was lost; 
    # a variable set to 'undef' or close of scope, because caller can't see it;
    # the current op has already been updated. The best we can do is indicate
    # 'near'.
-   warn "${\$self->__selfstr} was $self->{constructed_at} and was lost near $lost_at before it was ready.\n";
-} if DEBUG;
+
+   if( $self->{ready} and $self->{failure} ) {
+      warn "${\$self->__selfstr} was $self->{constructed_at} and was lost near $lost_at with an unreported failure of: " .
+         $self->{failure}[0] . "\n";
+   }
+   elsif( !$self->{ready} ) {
+      warn "${\$self->__selfstr} was $self->{constructed_at} and was lost near $lost_at before it was ready.\n";
+   }
+}
+*DESTROY = \&DESTROY_debug if DEBUG;
 
 =head2 $future = Future->done( @values )
 
@@ -352,6 +369,8 @@ sub _mark_ready
       next if $fail      and not( $flags & CB_FAIL );
       next if $cancelled and not( $flags & CB_CANCEL );
 
+      $self->{reported} = 1 if $fail;
+
       if( $is_future ) {
          $done ? $code->done( @result ) :
          $fail ? $code->fail( @result ) :
@@ -482,7 +501,8 @@ sub done
 Returns a C<CODE> reference that, when invoked, calls the C<done> method. This
 makes it simple to pass as a callback function to other code.
 
-The same effect can be achieved using L<curry>:
+As the same effect can be achieved using L<curry>, this method is deprecated
+now and may be removed in a later version.
 
  $code = $future->curry::done;
 
@@ -554,7 +574,8 @@ sub fail
 Returns a C<CODE> reference that, when invoked, calls the C<fail> method. This
 makes it simple to pass as a callback function to other code.
 
-The same effect can be achieved using L<curry>:
+As the same effect can be achieved using L<curry>, this method is deprecated
+now and may be removed in a later version.
 
  $code = $future->curry::fail;
 
@@ -692,13 +713,15 @@ sub on_ready
       my $fail = defined $self->{failure};
       my $done = !$fail && !$self->{cancelled};
 
+      $self->{reported} = 1 if $fail;
+
       $is_future ? ( $done ? $code->done( $self->get ) :
                      $fail ? $code->fail( $self->failure ) :
                              $code->cancel )
                  : $code->( $self );
    }
    else {
-      push @{ $self->{callbacks} }, [ CB_ALWAYS|CB_SELF, $code ];
+      push @{ $self->{callbacks} }, [ CB_ALWAYS|CB_SELF, $self->wrap_cb( on_ready => $code ) ];
    }
 
    return $self;
@@ -749,6 +772,7 @@ sub get
    my $self = shift;
    $self->await until $self->{ready};
    if( $self->{failure} ) {
+      $self->{reported} = 1;
       my $exception = $self->{failure}->[0];
       !ref $exception && $exception =~ m/\n$/ ? CORE::die $exception : Carp::croak $exception;
    }
@@ -821,7 +845,7 @@ sub on_done
                  : $code->( $self->get );
    }
    else {
-      push @{ $self->{callbacks} }, [ CB_DONE|CB_RESULT, $code ];
+      push @{ $self->{callbacks} }, [ CB_DONE|CB_RESULT, $self->wrap_cb( on_done => $code ) ];
    }
 
    return $self;
@@ -869,6 +893,7 @@ sub failure
    my $self = shift;
    $self->await until $self->{ready};
    return unless $self->{failure};
+   $self->{reported} = 1;
    return $self->{failure}->[0] if !wantarray;
    return @{ $self->{failure} };
 }
@@ -910,12 +935,13 @@ sub on_fail
 
    if( $self->{ready} ) {
       return $self if not $self->{failure};
+      $self->{reported} = 1;
 
       $is_future ? $code->fail( $self->failure )
                  : $code->( $self->failure );
    }
    else {
-      push @{ $self->{callbacks} }, [ CB_FAIL|CB_RESULT, $code ];
+      push @{ $self->{callbacks} }, [ CB_FAIL|CB_RESULT, $self->wrap_cb( on_fail => $code ) ];
    }
 
    return $self;
@@ -955,7 +981,8 @@ sub cancel
 Returns a C<CODE> reference that, when invoked, calls the C<cancel> method.
 This makes it simple to pass as a callback function to other code.
 
-The same effect can be achieved using L<curry>:
+As the same effect can be achieved using L<curry>, this method is deprecated
+now and may be removed in a later version.
 
  $code = $future->curry::cancel;
 
@@ -1040,6 +1067,10 @@ sub _sequence
    my $fseq = $f1->new;
    $fseq->on_cancel( $f1 );
 
+   # TODO: if anyone cares about the op name, we might have to synthesize it
+   # from $flags
+   $code = $f1->wrap_cb( sequence => $code ) unless $flags & (CB_SEQ_IMDONE|CB_SEQ_IMFAIL);
+
    push @{ $f1->{callbacks} }, [ CB_DONE|CB_FAIL|$flags, $code, $fseq ];
    weaken( $f1->{callbacks}[-1][2] );
 
@@ -1275,52 +1306,14 @@ sub followed_by
    return $self->_sequence( $code, CB_SEQ_ONDONE|CB_SEQ_ONFAIL|CB_SELF );
 }
 
-=head2 $future = $f1->and_then( \&code )
-
-An older form of C<then_with_f>; this method passes only the original future
-itself to the code, not its result. The code would have to call C<get> on the
-future to obtain the result.
-
- $f2 = $code->( $f1 )
-
-This method will be removed in a later version; use C<then_with_f> in new
-code. As a reminder, this method will now cause a warning every time it is
-invoked.
-
-=cut
-
 sub and_then
 {
-   my $self = shift;
-   my ( $code ) = @_;
-
-   Carp::carp "Future->and_then is now deprecated; use ->then_with_f instead";
-
-   return $self->_sequence( $code, CB_SEQ_ONDONE|CB_SELF );
+   Carp::croak "Future->and_then is now removed; use ->then_with_f instead";
 }
 
-=head2 $future = $f1->or_else( \&code )
-
-An older form of C<else_with_f>; this method passes only the original future
-itself to the code, not its failure and details. The code would have to call
-C<failure> on the future to obtain the result.
-
- $f2 = $code->( $f1 )
-
-This method will be removed in a later version; use C<else_with_f> in new
-code. As a reminder, this method will now cause a warning every time it is
-invoked.
-
-=cut
-
 sub or_else
 {
-   my $self = shift;
-   my ( $code ) = @_;
-
-   Carp::carp "Future->or_else is now deprecated; use ->else_with_f instead";
-
-   return $self->_sequence( $code, CB_SEQ_ONFAIL|CB_SELF );
+   Carp::croak "Future->or_else is now removed; use ->else_with_f instead";
 }
 
 =head2 $future = $f1->without_cancel
@@ -1675,7 +1668,7 @@ sub needs_any
 
    if( $immediate_done ) {
       foreach my $sub ( @subs ) {
-         $sub->{ready} or $sub->cancel;
+         $sub->{ready} ? $sub->{reported} = 1 : $sub->cancel;
       }
 
       $self->{result} = [ $immediate_done->get ];
@@ -1690,6 +1683,7 @@ sub needs_any
    }
 
    if( $immediate_fail ) {
+      $_->{reported} = 1 for @subs;
       # For consistency we'll pick the last one for the failure
       $self->{failure} = [ $subs[-1]->{failure} ];
       $self->_mark_ready( "needs_any" );
@@ -1863,6 +1857,60 @@ sub elapsed
    return $self->{elapsed} ||= tv_interval( $self->{btime}, $self->{rtime} );
 }
 
+=head2 $cb = $future->wrap_cb( $operation_name, $cb )
+
+I<Since version 0.31.>
+
+I<Note: This method is experimental and may be changed or removed in a later
+version.>
+
+This method is invoked internally by various methods that are about to save a
+callback CODE reference supplied by the user, to be invoked later. The default
+implementation simply returns the callback agument as-is; the method is
+provided to allow users to provide extra behaviour. This can be done by
+applying a method modifier of the C<around> kind, so in effect add a chain of
+wrappers. Each wrapper can then perform its own wrapping logic of the
+callback. C<$operation_name> is a string giving the reason for which the
+callback is being saved; currently one of C<on_ready>, C<on_done>, C<on_fail>
+or C<sequence>; the latter being used for all the sequence-returning methods.
+
+For example, the following wrapper saves the value of a package variable at
+the time the callback was saved, and restores that value at invocation time
+later on. This could be useful for preserving context during logging in a
+Future-based program.
+
+ our $LOGGING_CTX;
+
+ around "Future::wrap_cb" => sub {
+    my $orig = shift;
+    my $cb = $orig->( @_ );
+
+    my $saved_logging_ctx = $LOGGING_CTX;
+
+    return sub {
+       local $LOGGING_CTX = $saved_logging_ctx;
+       $cb->( @_ );
+    };
+ };
+
+At this point, any code deferred into a C<Future> by any of its callbacks will
+observe the C<$LOGGING_CTX> variable as having the value it held at the time
+the callback was saved, even if it is invoked later on when that value is
+different.
+
+Remember when writing such a wrapper, that it still needs to invoke the
+previous version of the method, so that it plays nicely in combination with
+others (see the C<< $orig->( @_ ) >> part).
+
+=cut
+
+sub wrap_cb
+{
+   my $self = shift;
+   my ( $op, $cb ) = @_;
+   return $cb;
+}
+
 =head1 EXAMPLES
 
 The following examples all demonstrate possible uses of a C<Future>
@@ -2038,9 +2086,9 @@ sequence future constructed from it.
 In particular, it is unclear in each of the following examples what the
 behaviour of C<$f2> should be, were C<$f1> to be cancelled:
 
- $f2 = $f1->then( sub { ... } ); # plus related ->then_with_f, ->and_then, ...
+ $f2 = $f1->then( sub { ... } ); # plus related ->then_with_f, ...
 
- $f2 = $f1->else( sub { ... } ); # plus related ->else_with_f, ->or_else, ...
+ $f2 = $f1->else( sub { ... } ); # plus related ->else_with_f, ...
 
  $f2 = $f1->followed_by( sub { ... } );
 
