@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use base qw( Test::Builder::Module );
 
-our $VERSION = '0.30';
+our $VERSION = '0.31';
 
 our @EXPORT = qw(
    no_pending_futures
@@ -72,11 +72,35 @@ sub no_pending_futures(&@)
 
    my @futures;
 
-   my $new = Future->can( "new" );
    no warnings 'redefine';
+
+   my $new = Future->can( "new" );
    local *Future::new = sub {
       my $f = $new->(@_);
       push @futures, $f;
+      $f->on_ready( sub {
+         my $f = shift;
+         for ( 0 .. $#futures ) {
+            refaddr( $futures[$_] ) == refaddr( $f ) or next;
+
+            splice @futures, $_, 1, ();
+            return;
+         }
+      });
+      return $f;
+   };
+
+   my $done = Future->can( "done" );
+   local *Future::done = sub {
+      my $f = $done->(@_);
+      pop @futures if !ref $_[0]; # class method
+      return $f;
+   };
+
+   my $fail = Future->can( "fail" );
+   local *Future::fail = sub {
+      my $f = $fail->(@_);
+      pop @futures if !ref $_[0]; # class method
       return $f;
    };
 
